module ContinuousNormalization where

open import Coinduction using (∞; ♯_; ♭)
open import Data.Nat using (ℕ)
open import Relation.Binary.PropositionalEquality using (_≡_)

infixl 9 _·_
infixr 5 _⟶_
infixl 5 _▻_
infix  5 _∙_ _∙-cong_
infix  1 _>>=_ _>>=-cong_
infix  0 _⊢_ _≈⊥_ _≈V_ _≈Env_ _≈⊢_

------------------------------------------------------------------------
-- A simply typed λ-calculus with a fixpoint operator

data Type : Set where
  nat : Type
  _⟶_ : (σ τ : Type) → Type

data Ctxt : Set where
  ε   : Ctxt
  _▻_ : (Γ : Ctxt) (τ : Type) → Ctxt

data _⊢_ : Ctxt → Type → Set where
  zero : ∀ {Γ τ} → Γ ▻ τ ⊢ τ
  wk   : ∀ {Γ σ τ} (t : Γ ⊢ τ) → Γ ▻ σ ⊢ τ
  ƛ    : ∀ {Γ σ τ} (t : Γ ▻ σ ⊢ τ) → Γ ⊢ σ ⟶ τ
  _·_  : ∀ {Γ σ τ} (t₁ : Γ ⊢ σ ⟶ τ) (t₂ : Γ ⊢ σ) → Γ ⊢ τ
  fix  : ∀ {Γ σ τ} (t₁ : Γ ⊢ (σ ⟶ τ) ⟶ (σ ⟶ τ)) (t₂ : Γ ⊢ σ) → Γ ⊢ τ

------------------------------------------------------------------------
-- Values

mutual

  data Env : Ctxt → Set where
    ε   : Env ε
    _▻_ : ∀ {Γ τ} (ρ : Env Γ) (v : Value τ) → Env (Γ ▻ τ)

  data Closure (σ τ : Type) : Set where
    ƛ : ∀ {Γ} (t : Γ ▻ σ ⊢ τ) (ρ : Env Γ) → Closure σ τ

  Value : Type → Set
  Value nat     = ℕ
  Value (σ ⟶ τ) = Closure σ τ

------------------------------------------------------------------------
-- The partiality monad

data _⊥ (A : Set) : Set where
  now   : (v : A) → A ⊥
  later : (x : ∞ (A ⊥)) → A ⊥

_>>=_ : {A B : Set} → A ⊥ → (A → B ⊥) → B ⊥
now v   >>= f = f v
later x >>= f = later (♯ (♭ x >>= f))

data Eq {A : Set} (_≈_ : A → A → Set) : A ⊥ → A ⊥ → Set where
  now    : ∀ {v₁ v₂} (v₁≈v₂ : v₁ ≈ v₂)                  → Eq _≈_ (now   v₁) (now   v₂)
  later  : ∀ {x₁ x₂} (x₁≈x₂ : ∞ (Eq _≈_ (♭ x₁) (♭ x₂))) → Eq _≈_ (later x₁) (later x₂)
  laterʳ : ∀ {x₁ x₂} (x₁≈x₂ :    Eq _≈_    x₁  (♭ x₂) ) → Eq _≈_        x₁  (later x₂)
  laterˡ : ∀ {x₁ x₂} (x₁≈x₂ :    Eq _≈_ (♭ x₁)    x₂  ) → Eq _≈_ (later x₁)        x₂

_>>=-cong_ :
  {A B : Set} {_≈A_ : A → A → Set} {_≈B_ : B → B → Set}
  {x₁ x₂ : A ⊥} {f₁ f₂ : A → B ⊥} →
  Eq _≈A_ x₁ x₂ → (∀ {v₁ v₂} → v₁ ≈A v₂ → Eq _≈B_ (f₁ v₁) (f₂ v₂)) →
  Eq _≈B_ (x₁ >>= f₁) (x₂ >>= f₂)
now    v₁≈v₂ >>=-cong f₁≈f₂ = f₁≈f₂ v₁≈v₂
later  x₁≈x₂ >>=-cong f₁≈f₂ = later (♯ (♭ x₁≈x₂ >>=-cong f₁≈f₂))
laterʳ x₁≈x₂ >>=-cong f₁≈f₂ = laterʳ   (  x₁≈x₂ >>=-cong f₁≈f₂)
laterˡ x₁≈x₂ >>=-cong f₁≈f₂ = laterˡ   (  x₁≈x₂ >>=-cong f₁≈f₂)

------------------------------------------------------------------------
-- Operational semantics

-- Not accepted by Agda, but my workaround works.

mutual

  ⟦_⟧ : ∀ {Γ τ} → Γ ⊢ τ → Env Γ → Value τ ⊥
  ⟦ zero      ⟧ (ρ ▻ v) = now v
  ⟦ wk t      ⟧ (ρ ▻ v) = ⟦ t ⟧ ρ
  ⟦ ƛ t       ⟧ ρ       = now (ƛ t ρ)
  ⟦ t₁ · t₂   ⟧ ρ       = ⟦ t₁ ⟧ ρ >>= λ f →
                          ⟦ t₂ ⟧ ρ >>= λ v →
                          f ∙ v
  ⟦ fix t₁ t₂ ⟧ ρ       = later (♯ ⟦ t₁ · ƛ (fix (wk t₁) zero) · t₂ ⟧ ρ)

  _∙_ : ∀ {σ τ} → Value (σ ⟶ τ) → Value σ → Value τ ⊥
  ƛ t₁ ρ ∙ v₂ = later (♯ ⟦ t₁ ⟧ (ρ ▻ v₂))

------------------------------------------------------------------------
-- Equality

-- Note the negative occurrence of _≈V_.

mutual

  _≈⊥_ : ∀ {τ} → Value τ ⊥ → Value τ ⊥ → Set
  _≈⊥_ = Eq _≈V_

  _≈V_ : ∀ {τ} → Value τ → Value τ → Set
  _≈V_ {nat}   v₁        v₂        = v₁ ≡ v₂
  _≈V_ {σ ⟶ τ} (ƛ t₁ ρ₁) (ƛ t₂ ρ₂) =
    ∀ {v₁ v₂} → v₁ ≈V v₂ → ∞ (⟦ t₁ ⟧ (ρ₁ ▻ v₁) ≈⊥ ⟦ t₂ ⟧ (ρ₂ ▻ v₂))

data _≈Env_ : ∀ {Γ} → Env Γ → Env Γ → Set where
  ε   : ε ≈Env ε
  _▻_ : ∀ {Γ τ} {v₁ v₂ : Value τ} {ρ₁ ρ₂ : Env Γ}
        (ρ₁≈ρ₂ : ρ₁ ≈Env ρ₂) (v₁≈v₂ : v₁ ≈V v₂) → ρ₁ ▻ v₁ ≈Env ρ₂ ▻ v₂

_≈⊢_ : ∀ {Γ τ} → Γ ⊢ τ → Γ ⊢ τ → Set
t₁ ≈⊢ t₂ = ∀ {ρ₁ ρ₂} → ρ₁ ≈Env ρ₂ → ⟦ t₁ ⟧ ρ₁ ≈⊥ ⟦ t₂ ⟧ ρ₂

------------------------------------------------------------------------
-- "Proof" of reflexivity

_∙-cong_ : ∀ {σ τ} {f₁ f₂ : Value (σ ⟶ τ)} {v₁ v₂ : Value σ} →
           f₁ ≈V f₂ → v₁ ≈V v₂ → f₁ ∙ v₁ ≈⊥ f₂ ∙ v₂
_∙-cong_ {f₁ = ƛ t₁ ρ₁} {f₂ = ƛ t₂ ρ₂} f₁≈f₂ v₁≈v₂ =  later (f₁≈f₂ v₁≈v₂)

-- Not accepted by Agda, and my workaround does not seem to work, due
-- to the negative occurrence of _≈V_ above. Can this definition be
-- handled using sized types?

refl⊢ : ∀ {Γ τ} (t : Γ ⊢ τ) → t ≈⊢ t
refl⊢ zero        (ρ₁≈ρ₂ ▻ v₁≈v₂) = now v₁≈v₂
refl⊢ (wk t)      (ρ₁≈ρ₂ ▻ v₁≈v₂) = refl⊢ t ρ₁≈ρ₂
refl⊢ (ƛ t)        ρ₁≈ρ₂          = now (λ v₁≈v₂ → ♯
                                           refl⊢ t (ρ₁≈ρ₂ ▻ v₁≈v₂))
refl⊢ (t₁ · t₂)    ρ₁≈ρ₂          = refl⊢ t₁ ρ₁≈ρ₂ >>=-cong λ f₁≈f₂ →
                                    refl⊢ t₂ ρ₁≈ρ₂ >>=-cong λ v₁≈v₂ →
                                    f₁≈f₂ ∙-cong v₁≈v₂
refl⊢ (fix t₁ t₂)  ρ₁≈ρ₂          =
  later (♯ refl⊢ (t₁ · ƛ (fix (wk t₁) zero) · t₂) ρ₁≈ρ₂)
