------------------------------------------------------------------------
-- An example which is non-trivial to define in Agda, but reasonably
-- straightforward (although tedious) using sized types in MiniAgda
--
-- Nils Anders Danielsson, with some help from Andreas Abel
------------------------------------------------------------------------

------------------------------------------------------------------------
-- Prelude

-- Natural numbers.

data Nat : Set
{ zero : Nat
; suc  : Nat -> Nat
}

-- Equality.

data Id (A : Set) (x : A) : A -> Set
{ refl : Id A x x
}

------------------------------------------------------------------------
-- The partiality monad

-- The partiality monad.

sized codata Lift (+ A : Set) : Size -> Set
{ now   : [i : Size] -> A        -> Lift A ($ i)
; later : [i : Size] -> Lift A i -> Lift A ($ i)
}

-- The bind operation of the monad.

fun bind : [A : Set] -> [B : Set] ->
           [i : Size] -> Lift A i -> (A -> Lift B i) -> Lift B i
{ bind A B ($ i) (now   .A .i a) f = f a
; bind A B ($ i) (later .A .i x) f = later B i (bind A B i x f)
}

-- The inductive part of weak bisimilarity.

data Eq' (A : Set) (+ Eq : Lift A # -> Lift A # -> Set)
         (+ eq : A -> A -> Set) :
         Lift A # -> Lift A # -> Set
{ nowCong'   : (v1 : A) -> (v2 : A) -> eq v1 v2 ->
               Eq' A Eq eq (now A # v1) (now A # v2)
; laterCong' : [x1 : Lift A #] -> [x2 : Lift A #] ->
               Eq x1 x2 -> Eq' A Eq eq (later A # x1) (later A # x2)
; laterL'    : [x1 : Lift A #] -> [x2 : Lift A #] ->
               Eq' A Eq eq x1 x2 -> Eq' A Eq eq (later A # x1) x2
; laterR'    : [x1 : Lift A #] -> [x2 : Lift A #] ->
               Eq' A Eq eq x1 x2 -> Eq' A Eq eq x1 (later A # x2)
}

-- Weak bisimilarity.

sized codata Eq (A : Set) (+ eq : A -> A -> Set) :
                Size -> Lift A # -> Lift A # -> Set
{ tie : [i : Size] -> [x1 : Lift A #] -> [x2 : Lift A #] ->
        Eq' A (Eq A eq i) eq x1 x2 -> Eq A eq ($ i) x1 x2
}

-- The constructors of Eq' can be "lifted" to the level of Eq.

fun nowCong : [A : Set] -> [eq : A -> A -> Set] ->
              [i : Size] -> (v1 : A) -> (v2 : A) ->
              eq v1 v2 -> Eq A eq i (now A # v1) (now A # v2)
{ nowCong A eq ($ i) v1 v2 vEq =
    tie A eq i (now A # v1) (now A # v2)
        (nowCong' A (Eq A eq i) eq v1 v2 vEq)
}

fun laterCong : [A : Set] -> [eq : A -> A -> Set] ->
                [i : Size] -> [x1 : Lift A #] -> [x2 : Lift A #] ->
                Eq A eq i x1 x2 ->
                Eq A eq ($ i) (later A # x1) (later A # x2)
{ laterCong A eq i x1 x2 xEq =
    tie A eq i (later A # x1) (later A # x2)
        (laterCong' A (Eq A eq i) eq x1 x2 xEq)
}

fun laterL : [A : Set] -> [eq : A -> A -> Set] ->
             [i : Size] -> [x1 : Lift A #] -> [x2 : Lift A #] ->
             Eq A eq i x1 x2 -> Eq A eq i (later A # x1) x2
{ laterL A eq ($ i) x1 x2 (tie .A .eq .i .x1 .x2 xEq) =
    tie A eq i (later A # x1) x2
        (laterL' A (Eq A eq i) eq x1 x2 xEq)
}

fun laterR : [A : Set] -> [eq : A -> A -> Set] ->
             [i : Size] -> [x1 : Lift A #] -> [x2 : Lift A #] ->
             Eq A eq i x1 x2 -> Eq A eq i x1 (later A # x2)
{ laterR A eq ($ i) x1 x2 (tie .A .eq .i .x1 .x2 xEq) =
    tie A eq i x1 (later A # x2)
        (laterR' A (Eq A eq i) eq x1 x2 xEq)
}

-- Bind preserves weak bisimilarity.

mutual {

  fun bindCong' : [i : Size] -> [A : Set] -> [B : Set] ->
                  [eqA : A -> A -> Set] -> [eqB : B -> B -> Set] ->
                  [x1 : Lift A #] -> [x2 : Lift A #] ->
                  [f1 : A -> Lift B #] -> [f2 : A -> Lift B #] ->
                  Eq' A (Eq A eqA i) eqA x1 x2 ->
                  ((v1 : A) -> (v2 : A) -> eqA v1 v2 ->
                   Eq B eqB ($ i) (f1 v1) (f2 v2)) ->
                  Eq B eqB ($ i) (bind A B # x1 f1) (bind A B # x2 f2)
  { bindCong' i A B eqA eqB .(now A # v1) .(now A # v2) f1 f2
      (nowCong' .A .(Eq A eqA i) .eqA v1 v2 vEq) fEq =
      fEq v1 v2 vEq
  ; bindCong' i A B eqA eqB .(later A # x1) .(later A # x2) f1 f2
      (laterCong' .A .(Eq A eqA i) .eqA x1 x2 xEq) fEq =
      laterCong B eqB i (bind A B # x1 f1) (bind A B # x2 f2)
                (bindCong i A B eqA eqB x1 x2 f1 f2 xEq fEq)
  ; bindCong' i .A B eqA eqB .(later A # x1) .x2 f1 f2
      (laterL' A .(Eq A eqA i) .eqA x1 x2 xEq) fEq =
      laterL B eqB ($ i) (bind A B # x1 f1) (bind A B # x2 f2)
             (bindCong' i A B eqA eqB x1 x2 f1 f2 xEq fEq)
  ; bindCong' i .A B eqA eqB .x1 .(later A # x2) f1 f2
      (laterR' A .(Eq A eqA i) .eqA x1 x2 xEq) fEq =
      laterR B eqB ($ i) (bind A B # x1 f1) (bind A B # x2 f2)
             (bindCong' i A B eqA eqB x1 x2 f1 f2 xEq fEq)
  }

  fun bindCong : [i : Size] -> [A : Set] -> [B : Set] ->
                 [eqA : A -> A -> Set] -> [eqB : B -> B -> Set] ->
                 [x1 : Lift A #] -> [x2 : Lift A #] ->
                 [f1 : A -> Lift B #] -> [f2 : A -> Lift B #] ->
                 Eq A eqA i x1 x2 ->
                 ((v1 : A) -> (v2 : A) -> eqA v1 v2 -> Eq B eqB i (f1 v1) (f2 v2)) ->
                 Eq B eqB i (bind A B # x1 f1) (bind A B # x2 f2)
  { bindCong ($ i) A B eqA eqB x1 x2 f1 f2
             (tie .A .eqA .i .x1 .x2 xEq) fEq =
      bindCong' i A B eqA eqB x1 x2 f1 f2 xEq fEq
  }

}

------------------------------------------------------------------------
-- A simply typed lambda-calculus with a fixpoint operator

-- Types.

data Type : Set
{ nat : Type
; arr : Type -> Type -> Type
}

-- Contexts.

data Ctxt : Set
{ empty : Ctxt
; snoc  : Ctxt -> Type -> Ctxt
}

-- Well-typed terms.

data Tm : Ctxt -> Type -> Set
{ vz  : [G : Ctxt] ->               [b : Type] -> Tm (snoc G b) b
; wk  : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm G b -> Tm (snoc G a) b
; lam : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm (snoc G a) b -> Tm G (arr a b)
; app : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm G (arr a b) -> Tm G a -> Tm G b
; fix : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm G (arr (arr a b) (arr a b)) -> Tm G a -> Tm G b
}

------------------------------------------------------------------------
-- Environments and values

-- Environments and values are defined as a single family, indexed by
-- a "flag".

data Flag : Set
{ env : Flag
; val : Flag
}

-- The environment-value family has another index as well, given by
-- the following type.

fun Index : Flag -> Set
{ Index env = Ctxt
; Index val = Type
}

-- The environment-value family.

data EnvOrVal : (f : Flag) -> Index f -> Set
{ emptyE : EnvOrVal env empty
; snocE  : [G : Ctxt] -> [b : Type] -> EnvOrVal env G -> EnvOrVal val b -> EnvOrVal env (snoc G b)
; natV   : Nat -> EnvOrVal val nat
; arrV   : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm (snoc G a) b -> EnvOrVal env G -> EnvOrVal val (arr a b)
}

-- Environments.

fun Env : Ctxt -> Set
{ Env = EnvOrVal env
}

-- Values.

fun Value : Type -> Set
{ Value = EnvOrVal val
}

------------------------------------------------------------------------
-- Operational semantics

-- A helper function, used to define the semantics of the fixpoint
-- operator.

fun unfoldedFixpoint : [G : Ctxt] -> [a : Type] -> [b : Type] ->
                       Tm G (arr (arr a b) (arr a b)) -> Tm G a -> Tm G b
{ unfoldedFixpoint G a b t1 t2 =
    app G a b
      (app G (arr a b) (arr a b)
         t1
         (lam G a b
            (fix (snoc G a) a b
               (wk G a (arr (arr a b) (arr a b)) t1)
               (vz G a))))
      t2
}

mutual {

  -- Operational semantics.

  fun sem : [i : Size] -> [G : Ctxt] -> [b : Type] ->
            Tm G b -> Env G -> Lift (Value b) i
  { sem i     .(snoc G b) .b         (vz G b)          (snocE .G .b rho v) = now (Value b) i v
  ; sem i     .(snoc G a) .b         (wk G a b t)      (snocE .G .a rho v) = sem i G b t rho
  ; sem i     .G          .(arr a b) (lam G a b t)     rho                 = now (Value (arr a b)) i (arrV G a b t rho)
  ; sem i     .G          .b         (app G a b t1 t2) rho =
     bind (Value (arr a b)) (Value b) i (sem i G (arr a b) t1 rho) (\ f ->
     bind (Value a)         (Value b) i (sem i G a         t2 rho) (\ v ->
     appV i a b f v))
  ; sem ($ i) .G          .b         (fix G a b t1 t2) rho =
     later (Value b) i (sem i G b (unfoldedFixpoint G a b t1 t2) rho)
  }

  -- Value application.

  fun appV : [i : Size] -> [a : Type] -> [b : Type] ->
             Value (arr a b) -> Value a -> Lift (Value b) i
  { appV ($ i) .a .b (arrV G a b t1 rho) v2 =
     later (Value b) i (sem i (snoc G a) b t1 (snocE G a rho v2))
  }

}

------------------------------------------------------------------------
-- Equality

mutual {

  -- Equality of computations.

  fun EqL : -Size -> (a : Type) -> Lift (Value a) # -> Lift (Value a) # -> Set
  { EqL i a x y = Eq (Value a) (EqV i a) i x y
  }

  -- Equality of values.

  fun EqV : -Size -> (a : Type) -> Value a -> Value a -> Set
  { EqV i nat       (natV x)                (natV y)                = Id Nat x y
  ; EqV i (arr a b) (arrV G1 .a .b t1 rho1) (arrV G2 .a .b t2 rho2) =
      (v1 : Value a) -> (v2 : Value a) ->
      EqV # a v1 v2 ->
      EqL i b (sem # (snoc G1 a) b t1 (snocE G1 a rho1 v1))
              (sem # (snoc G2 a) b t2 (snocE G2 a rho2 v2))
  }

}

-- Equality of environments.

data EqEnv : (G : Ctxt) -> Env G -> Env G -> Set
{ emptyEq : EqEnv empty emptyE emptyE
; snocEq  : [G : Ctxt] -> [b : Type] ->
            [rho1 : Env G] -> [rho2 : Env G] ->
            [v1 : Value b] -> [v2 : Value b] ->
            EqEnv G rho1 rho2 -> EqV # b v1 v2 ->
            EqEnv (snoc G b) (snocE G b rho1 v1) (snocE G b rho2 v2)
}

-- Equality of terms.

fun EqT : (G : Ctxt) -> (b : Type) -> Tm G b -> Tm G b -> Set
{ EqT G b t1 t2 = [i : Size] -> (rho1 : Env G) -> (rho2 : Env G) ->
                  EqEnv G rho1 rho2 ->
                  EqL i b (sem # G b t1 rho1) (sem # G b t2 rho2)
}

------------------------------------------------------------------------
-- Some lemmas

-- Value application preserves equality.

fun appCong : [a : Type] -> [b : Type] ->
              [f1 : Value (arr a b)] -> [f2 : Value (arr a b)] ->
              (v1 : Value a) -> (v2 : Value a) ->
              EqV # (arr a b) f1 f2 -> EqV # a v1 v2 ->
              EqL # b (appV # a b f1 v1) (appV # a b f2 v2)
{ appCong a b (arrV G1 .a .b t1 rho1) (arrV G2 .a .b t2 rho2) v1 v2 f1f2 v1v2 =
    laterCong (Value b) (EqV # b) #
              (sem # (snoc G1 a) b t1 (snocE G1 a rho1 v1))
              (sem # (snoc G2 a) b t2 (snocE G2 a rho2 v2))
              (f1f2 v1 v2 v1v2)
}

-- The term equality is reflexive.

fun reflT : [G : Ctxt] -> [b : Type] -> (t : Tm G b) -> EqT G b t t
{ reflT .(snoc G b) .b (vz G b) i
        (snocE .G .b rho1 v1)
        (snocE .G .b rho2 v2)
        (snocEq .G .b .rho1 .rho2 .v1 .v2 rhoEq vEq) =
    nowCong (Value b) (EqV # b) # v1 v2 vEq
; reflT .(snoc G a) .b (wk G a b t) i
        (snocE .G .a rho1 v1)
        (snocE .G .a rho2 v2)
        (snocEq .G .a .rho1 .rho2 .v1 .v2 rhoEq vEq) =
    reflT G b t i rho1 rho2 rhoEq
; reflT .G .(arr a b) (lam G a b t) i rho1 rho2 rhoEq =
    nowCong (Value (arr a b)) (EqV i (arr a b)) #
            (arrV G a b t rho1)
            (arrV G a b t rho2)
      (\ v1 -> \ v2 -> \ vEq ->
         reflT (snoc G a) b t i
               (snocE G a rho1 v1)
               (snocE G a rho2 v2)
               (snocEq G a rho1 rho2 v1 v2 rhoEq vEq))
; reflT .G .b (app G a b t1 t2) i rho1 rho2 rhoEq =
   bindCong i (Value (arr a b)) (Value b) (EqV # (arr a b)) (EqV # b)
            (sem # G (arr a b) t1 rho1)
            (sem # G (arr a b) t1 rho2)
            (\ f -> bind (Value a) (Value b) # (sem # G a t2 rho1) (\ v -> appV # a b f v))
            (\ f -> bind (Value a) (Value b) # (sem # G a t2 rho2) (\ v -> appV # a b f v))
            (reflT G (arr a b) t1 i rho1 rho2 rhoEq)
            (\ f1 -> \ f2 -> \ fEq ->
   bindCong i (Value a) (Value b) (EqV # a) (EqV # b)
            (sem # G a t2 rho1)
            (sem # G a t2 rho2)
            (appV # a b f1)
            (appV # a b f2)
            (reflT G a t2 i rho1 rho2 rhoEq)
            (\ v1 -> \ v2 -> \ vEq ->
   appCong a b f1 f2 v1 v2 fEq vEq))
; reflT .G .b (fix G a b t1 t2) ($ i) rho1 rho2 rhoEq =
    let t : Tm G b = unfoldedFixpoint G a b t1 t2 in
    laterCong (Value b) (EqV # b) i
              (sem # G b t rho1)
              (sem # G b t rho2)
              (reflT G b t i rho1 rho2 rhoEq)
}
